@@ -10,17 +10,21 @@ SEXP installAttrib(SEXP vec, SEXP name, SEXP val)
     SEXP t = R_NilValue; /* -Wall */
 
     if(TYPEOF(vec) == CHARSXP)
-	error("cannot set attribute on a CHARSXP");
+	error("cannot set attribute on a CHARSXP"); 		/* #nocov */
+    if (TYPEOF(vec) == SYMSXP)
+	error("cannot set attribute on a symbol"); 		/* #nocov */
     /* this does no allocation */
     for (SEXP s = ATTRIB(vec); s != R_NilValue; s = CDR(s)) {
 	if (TAG(s) == name) {
-	    SETCAR(s, val);
-	    return val;
+	    SETCAR(s, val); 					/* #nocov */
+	    return val; 					/* #nocov */
 	}
 	t = s; // record last attribute, if any
     }
-    /* The usual convention is that the caller protects, 
-       so this is historical over-cautiousness */
+
+    /* The usual convention is that the caller protects,
+       but a lot of existing code depends assume that
+       setAttrib/installAttrib protects its arguments */
     PROTECT(vec); PROTECT(name); PROTECT(val);
     SEXP s = CONS(val, R_NilValue);
     SET_TAG(s, name);
