@@ -1,6 +1,6 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
- *  Copyright (C) 2000-2016  The R Core Team.
+ *  Copyright (C) 2000-2017  The R Core Team.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -39,10 +39,18 @@
 
 #include <Rmath.h> // Rexp10
 
-// some other header, e.g. math.h, might define it
-#if defined(__GLIBC__) && !defined(_BSD_SOURCE)
-// to get tm_zone, tm_gmtoff defined
-# define _BSD_SOURCE
+// to get tm_zone, tm_gmtoff defined in glibc.
+// some other header, e.g. math.h, might define the macro.
+#if defined HAVE_FEATURES_H
+# include <features.h>
+# ifdef __GNUC_PREREQ
+#  if __GNUC_PREREQ(2,20) && !defined(_DEFAULT_SOURCE_)
+#   define _DEFAULT_SOURCE 1
+#  endif
+# endif
+#endif
+#if defined(HAVE_GLIBC2) && !defined(_DEFAULT_SOURCE_) && !defined(_BSD_SOURCE)
+# define _BSD_SOURCE 1
 #endif
 #include <time.h>
 
@@ -133,51 +141,51 @@ static int validate_tm (stm *tm)
     int tmp, res = 0;
 
     if (tm->tm_sec < 0 || tm->tm_sec > 60) { /* 61 POSIX, 60 draft ISO C */
-	res++;
+	res++; 							/* #nocov start */
 	tmp = tm->tm_sec/60;
 	tm->tm_sec -= 60 * tmp; tm->tm_min += tmp;
 	if(tm->tm_sec < 0) {tm->tm_sec += 60; tm->tm_min--;}
     }
 
     if (tm->tm_min < 0 || tm->tm_min > 59) {
-	res++;
+	res++; 							/* #nocov start */
 	tmp = tm->tm_min/60;
 	tm->tm_min -= 60 * tmp; tm->tm_hour += tmp;
-	if(tm->tm_min < 0) {tm->tm_min += 60; tm->tm_hour--;}
+	if(tm->tm_min < 0) {tm->tm_min += 60; tm->tm_hour--;}	/* #nocov end */
     }
 
     if(tm->tm_hour == 24 && tm->tm_min == 0 && tm->tm_sec == 0) {
-	tm->tm_hour = 0; tm->tm_mday++;
+	tm->tm_hour = 0; tm->tm_mday++; 			/* #nocov start */
 	if(tm->tm_mon >= 0 && tm->tm_mon <= 11) {
 	    if(tm->tm_mday > days_in_month[tm->tm_mon] +
 	       ((tm->tm_mon==1 && isleap(1900+tm->tm_year) ? 1 : 0))) {
 		   tm->tm_mon++; tm->tm_mday = 1;
 		   if(tm->tm_mon == 12) {
 		       tm->tm_year++; tm->tm_mon = 0;
-		   }
+		   }						/* #nocov end */
 	       }
 	}
     }
     if (tm->tm_hour < 0 || tm->tm_hour > 23) {
-	res++;
+	res++; 							/* #nocov start */
 	tmp = tm->tm_hour/24;
 	tm->tm_hour -= 24 * tmp; tm->tm_mday += tmp;
-	if(tm->tm_hour < 0) {tm->tm_hour += 24; tm->tm_mday--;}
+	if(tm->tm_hour < 0) {tm->tm_hour += 24; tm->tm_mday--;}	/* #nocov end */
     }
 
     /* defer fixing mday until we know the year */
     if (tm->tm_mon < 0 || tm->tm_mon > 11) {
-	res++;
+	res++; 							/* #nocov start */
 	tmp = tm->tm_mon/12;
 	tm->tm_mon -= 12 * tmp; tm->tm_year += tmp;
-	if(tm->tm_mon < 0) {tm->tm_mon += 12; tm->tm_year--;}
+	if(tm->tm_mon < 0) {tm->tm_mon += 12; tm->tm_year--;} 	/* #nocov end */
     }
 
     /* A limit on the loops of about 3000x round */
     if(tm->tm_mday < -1000000 || tm->tm_mday > 1000000) return -1;
 
     if(abs(tm->tm_mday) > 366) {
-	res++;
+	res++; 							/* #nocov start */
 	/* first spin back until January */
 	while(tm->tm_mon > 0) {
 	    --tm->tm_mon;
@@ -192,22 +200,22 @@ static int validate_tm (stm *tm)
 	while(tm->tm_mday >
 	      (tmp = 365 + (isleap(1900+tm->tm_year)? 1 : 0))) {
 	    tm->tm_mday -= tmp; tm->tm_year++;
-	}
+	} 							/* #nocov end */
     }
 
     while(tm->tm_mday < 1) {
-	res++;
+	res++; 							/* #nocov start */
 	if(--tm->tm_mon < 0) {tm->tm_mon += 12; tm->tm_year--;}
 	tm->tm_mday += days_in_month[tm->tm_mon] +
 	    ((tm->tm_mon==1 && isleap(1900+tm->tm_year))? 1 : 0);
-    }
+    } 								/* #nocov end */
 
     while(tm->tm_mday >
 	  (tmp = days_in_month[tm->tm_mon] +
 	   ((tm->tm_mon==1 && isleap(1900+tm->tm_year))? 1 : 0))) {
-	res++;
+	res++; 							/* #nocov start */
 	if(++tm->tm_mon > 11) {tm->tm_mon -= 12; tm->tm_year++;}
-	tm->tm_mday -= tmp;
+	tm->tm_mday -= tmp; 					/* #nocov end */
     }
     return res;
 }
@@ -223,13 +231,13 @@ static double mktime00 (stm *tm)
     day = tm->tm_mday - 1;
     year0 = 1900 + tm->tm_year;
     /* safety check for unbounded loops */
-    if (year0 > 3000) {
+    if (year0 > 3000) {						/* #nocov start */
 	excess = (int)(year0/2000) - 1;
 	year0 -= (int)(excess * 2000);
     } else if (year0 < 0) {
 	excess = -1 - (int)(-year0/2000);
 	year0 -= (int)(excess * 2000);
-    }
+    }								/* #nocov end */
 
     for(int i = 0; i < tm->tm_mon; i++) day += days_in_month[i];
     if (tm->tm_mon > 1 && isleap(year0)) day++;
@@ -314,7 +322,7 @@ static const time_t leapseconds[] = // dput(unclass(.leap.seconds)) :
    915148800,1136073600,1230768000,1341100800,1435708800,1483228800};
 #endif
 
-static double guess_offset (stm *tm)
+static double guess_offset (stm *tm)					/* #nocov start */ 
 {
     double offset, offset1, offset2;
     int i, wday, year, oldmonth, oldisdst, oldmday;
@@ -431,7 +439,7 @@ static double mktime0 (stm *tm, const int local)
 	return res;
 /* watch the side effect here: both calls alter their arg */
     } else return guess_offset(tm) + mktime00(tm);
-}
+}								/* #nocov end */ 
 
 /* Interface for localtime or gmtime or internal substitute */
 static stm * localtime0(const double *tp, const int local, stm *ltm)
@@ -467,7 +475,7 @@ static stm * localtime0(const double *tp, const int local, stm *ltm)
 
     /* internal substitute code.
        Like localtime, this returns a pointer to a static struct tm */
-
+									/* #nocov start */ 
     int day = (int) floor(d/86400.0);
     int left = (int) (d - day * 86400.0 + 1e-6); // allow for fractional secs
 
@@ -539,7 +547,7 @@ static stm * localtime0(const double *tp, const int local, stm *ltm)
     } else {
 	res->tm_isdst = 0; /* no dst in GMT */
 	return res;
-    }
+    }								/* #nocov end */ 
 }
 #endif
 
@@ -552,7 +560,7 @@ static int set_tz(const char *tz, char *oldtz)
     p = getenv("TZ");
     if(p) {
 	if (strlen(p) > 1000)
-	    error("time zone specification is too long");
+	    error("time zone specification is too long");	/* #nocov */
 	strcpy(oldtz, p);
     }
 #ifdef HAVE_SETENV
@@ -562,7 +570,7 @@ static int set_tz(const char *tz, char *oldtz)
     {
 	static char buff[1010];
 	if (strlen(tz) > 1000)
-	    error("time zone specification is too long");
+	    error("time zone specification is too long");	/* #nocov */
 	strcpy(buff, "TZ="); strcat(buff, tz);
 	if(putenv(buff)) warning("problem with setting timezone");
     }
@@ -598,7 +606,7 @@ static void reset_tz(char *tz)
     tzset();
 }
 
-static void glibc_fix(stm *tm, int *invalid)
+static void glibc_fix(stm *tm, int *invalid) 			/* #nocov start */ 
 {
     /* set mon and mday which glibc does not always set.
        Use current year/... if none has been specified.
@@ -639,7 +647,7 @@ static void glibc_fix(stm *tm, int *invalid)
 	}
 	if(tm->tm_mon == NA_INTEGER) tm->tm_mon = tm0->tm_mon;
     }
-}
+}								/* #nocov end */ 
 
 
 static const char ltnames [][7] =
@@ -661,10 +669,10 @@ makelt(stm *tm, SEXP ans, R_xlen_t i, int valid, double frac_secs)
 	INTEGER(VECTOR_ELT(ans, 7))[i] = tm->tm_yday;
 	INTEGER(VECTOR_ELT(ans, 8))[i] = tm->tm_isdst;
     } else {
-	REAL(VECTOR_ELT(ans, 0))[i] = NA_REAL;
+	REAL(VECTOR_ELT(ans, 0))[i] = NA_REAL;			/* #nocov start */
 	for(int j = 1; j < 8; j++)
 	    INTEGER(VECTOR_ELT(ans, j))[i] = NA_INTEGER;
-	INTEGER(VECTOR_ELT(ans, 8))[i] = -1;
+	INTEGER(VECTOR_ELT(ans, 8))[i] = -1;			/* #nocov end */
     }
 }
 
@@ -687,7 +695,7 @@ SEXP asPOSIXlt(SEXP argsxp, SEXP tzarg) // other args?
     /*     error("invalid '%s' value", "tz"); */
     /* tz = CHAR(STRING_ELT(stz, 0)); */
     if (!isString((stz = tzarg)) || LENGTH(stz) != 1)
-        error("invalid '%s' value", "tz");
+        error("invalid '%s' value", "tz");			/* #nocov */
     tz = CHAR(STRING_ELT(stz, 0));
     if(strlen(tz) == 0) {
 	/* do a direct look up here as this does not otherwise
@@ -709,7 +717,7 @@ SEXP asPOSIXlt(SEXP argsxp, SEXP tzarg) // other args?
 
     // localtime may change tzname.
     if (isgmt) {
-	PROTECT(tzone = mkString(tz));
+	PROTECT(tzone = mkString(tz));				/* #nocov */
     } else {
 	PROTECT(tzone = allocVector(STRSXP, 3));
 	SET_STRING_ELT(tzone, 0, mkChar(tz));
@@ -746,7 +754,7 @@ SEXP asPOSIXlt(SEXP argsxp, SEXP tzarg) // other args?
 	       struct tm pointer, but Windows uses NULL for error
 	       conditions (like negative times). */
 	    valid = (ptm != NULL);
-	} else valid = 0;
+	} else valid = 0;				/* #nocov */
 	makelt(ptm, ans, i, valid, d - floor(d));
 	if(!isgmt) {
 	    char *p = "";
@@ -788,10 +796,10 @@ SEXP asPOSIXct(SEXP sxparg, SEXP tzarg) //
     /* PROTECT(x = duplicate(CAR(args))); /\* coerced below *\/ */
     PROTECT(x = duplicate(sxparg)); /* coerced below */ 
     if (!isVectorList(x) || LENGTH(x) < 9)
-        error("invalid '%s' argument", "x");
+        error("invalid '%s' argument", "x");			/* #nocov */
     /* if(!isString((stz = CADR(args))) || LENGTH(stz) != 1) */
     if (!isString((stz = tzarg)) || LENGTH(stz) != 1)
-        error("invalid '%s' value", "tz");
+        error("invalid '%s' value", "tz");			/* #nocov */
 
     tz = CHAR(STRING_ELT(stz, 0));
     if(strlen(tz) == 0) {
@@ -818,9 +826,9 @@ SEXP asPOSIXct(SEXP sxparg, SEXP tzarg) //
     if(n > 0) {
 	for(int i = 0; i < 6; i++)
 	    if(nlen[i] == 0)
-		error("zero-length component in non-empty \"POSIXlt\" structure");
+		error("zero-length component in non-empty \"POSIXlt\" structure"); /* #nocov */
 	if(nlen[8] == 0)
-	    error("zero-length component in non-empty \"POSIXlt\" structure");
+	    error("zero-length component in non-empty \"POSIXlt\" structure");	   /* #nocov */
     }
     /* coerce fields to integer or real */
     SET_VECTOR_ELT(x, 0, coerceVector(VECTOR_ELT(x, 0), REALSXP));
@@ -844,7 +852,7 @@ SEXP asPOSIXct(SEXP sxparg, SEXP tzarg) //
 	if(!R_FINITE(secs) || tm.tm_min == NA_INTEGER ||
 	   tm.tm_hour == NA_INTEGER || tm.tm_mday == NA_INTEGER ||
 	   tm.tm_mon == NA_INTEGER || tm.tm_year == NA_INTEGER)
-	    REAL(ans)[i] = NA_REAL;
+	    REAL(ans)[i] = NA_REAL;					/* #nocov */
 	else {
 	    errno = 0;
 	    tmp = mktime0(&tm, 1 - isgmt);
@@ -853,8 +861,8 @@ SEXP asPOSIXct(SEXP sxparg, SEXP tzarg) //
 #else
 	    REAL(ans)[i] = ((tmp == -1.)
 			    /* avoid silly gotcha at epoch minus one sec */
-			    && (tm.tm_sec != 59)
-			    && ((tm.tm_sec = 58), (mktime0(&tm, 1 - isgmt) != -2.))
+			    && (tm.tm_sec != 59)					/* #nocov */
+			    && ((tm.tm_sec = 58), (mktime0(&tm, 1 - isgmt) != -2.))	/* #nocov */
 			    ) ?
 	      NA_REAL : tmp + (secs - fsecs);
 #endif
@@ -887,15 +895,15 @@ SEXP formatPOSIXlt(SEXP argsxp, SEXP fmtsxp, SEXP tzsxp) //
     /* PROTECT(x = duplicate(CAR(args))); /\* coerced below *\/ */
     PROTECT(x = duplicate(argsxp)); /* coerced below */ 
     if (!isVectorList(x) || LENGTH(x) < 9)
-        error("invalid '%s' argument", "x");
+        error("invalid '%s' argument", "x");					/* #nocov */
     /* if(!isString((sformat = CADR(args))) || XLENGTH(sformat) == 0) */
     if (!isString((sformat = fmtsxp)) || XLENGTH(sformat) == 0)
-        error("invalid '%s' argument", "format");
+        error("invalid '%s' argument", "format");				/* #nocov */
     m = XLENGTH(sformat);
     /* UseTZ = asLogical(CADDR(args)); */
     UseTZ = asLogical(tzsxp);
     if (UseTZ == NA_LOGICAL)
-	error("invalid '%s' argument", "usetz");
+	error("invalid '%s' argument", "usetz");				/* #nocov */
     tz = getAttrib(x, install("tzone"));
 
     const char *tz1;
@@ -926,12 +934,19 @@ SEXP formatPOSIXlt(SEXP argsxp, SEXP fmtsxp, SEXP tzsxp) //
     if(n > 0) {
 	for(int i = 0; i < 9; i++)
 	    if(nlen[i] == 0)
-		error("zero-length component in non-empty \"POSIXlt\" structure");
+		error("zero-length component in non-empty \"POSIXlt\" structure");	/* #nocov */
     }
     if(n > 0) N = (m > n) ? m:n; else N = 0;
     PROTECT(ans = allocVector(STRSXP, N));
     char tm_zone[20];
-    Rboolean have_zone = LENGTH(x) >= 10 && LENGTH(VECTOR_ELT(x, 9)) == n;
+#ifdef HAVE_TM_GMTOFF
+    Rboolean have_zone = LENGTH(x) >= 11 && XLENGTH(VECTOR_ELT(x, 9)) == n &&
+	XLENGTH(VECTOR_ELT(x, 10)) == n;
+#else
+    Rboolean have_zone = LENGTH(x) >= 10 && XLENGTH(VECTOR_ELT(x, 9)) == n;
+#endif
+    if(have_zone && !isString(VECTOR_ELT(x, 9)))
+	error("invalid component [[10]] in \"POSIXlt\" should be 'zone'");		/* #nocov */
     for(R_xlen_t i = 0; i < N; i++) {
 	double secs = REAL(VECTOR_ELT(x, 0))[i%nlen[0]], fsecs = floor(secs);
 	// avoid (int) NAN
@@ -963,75 +978,79 @@ SEXP formatPOSIXlt(SEXP argsxp, SEXP fmtsxp, SEXP tzsxp) //
 	if(!R_FINITE(secs) || tm.tm_min == NA_INTEGER ||
 	   tm.tm_hour == NA_INTEGER || tm.tm_mday == NA_INTEGER ||
 	   tm.tm_mon == NA_INTEGER || tm.tm_year == NA_INTEGER) {
-	    SET_STRING_ELT(ans, i, NA_STRING);
+	    SET_STRING_ELT(ans, i, NA_STRING);					/* #nocov */
+	} else if(validate_tm(&tm) < 0) {
+	    SET_STRING_ELT(ans, i, NA_STRING);					/* #nocov */
 	} else {
-	    if(validate_tm(&tm) < 0) SET_STRING_ELT(ans, i, NA_STRING);
-	    else {
-		const char *q = translateChar(STRING_ELT(sformat, i%m));
-		int n = (int) strlen(q) + 50;
-		char buf2[n];
-		const char *p;
+	    const char *q = translateChar(STRING_ELT(sformat, i%m));
+	    int nn = (int) strlen(q) + 50;
+	    char buf2[nn];
+	    const char *p;
 #ifdef OLD_Win32
-		/* We want to override Windows' TZ names */
-		p = strstr(q, "%Z");
-		if (p) {
-		    memset(buf2, 0, n);
-		    strncpy(buf2, q, p - q);
-		    if(have_zone) 
-			strcat(buf2, tm_zone);
-		    else
-			strcat(buf2, tm.tm_isdst > 0 ? R_tzname[1] : R_tzname[0]);
-		    strcat(buf2, p+2);
-		} else
+	    /* We want to override Windows' TZ names */
+	    p = strstr(q, "%Z");
+	    if (p) {
+		memset(buf2, 0, nn);
+		strncpy(buf2, q, p - q);
+		if(have_zone)
+		    strcat(buf2, tm_zone);
+		else
+		    strcat(buf2, tm.tm_isdst > 0 ? R_tzname[1] : R_tzname[0]);
+		strcat(buf2, p+2);
+	    } else
 #endif
-		    strcpy(buf2, q);
+		strcpy(buf2, q);
 
-		p = strstr(q, "%OS");
-		if(p) {
-		    /* FIXME some of this should be outside the loop */
-		    int ns, nused = 4;
-		    char *p2 = strstr(buf2, "%OS");
-		    *p2 = '\0';
-		    ns = *(p+3) - '0';
-		    if(ns < 0 || ns > 9) { /* not a digit */
-			ns = asInteger(GetOption1(install("digits.secs")));
-			if(ns == NA_INTEGER) ns = 0;
-			nused = 3;
-		    }
-		    if(ns > 6) ns = 6;
-		    if(ns > 0) {
-			/* truncate to avoid nuisances such as PR#14579 */
-			double s = secs, t = Rexp10((double) ns);
-			s = ((int) (s*t))/t;
-			sprintf(p2, "%0*.*f", ns+3, ns, s);
-			strcat(buf2, p+nused);
-		    } else {
-			strcat(p2, "%S");
-			strcat(buf2, p+nused);
-		    }
+	    p = strstr(q, "%OS");
+	    if(p) {
+		/* FIXME some of this should be outside the loop */
+		int ns, nused = 4;
+		char *p2 = strstr(buf2, "%OS");
+		*p2 = '\0';
+		ns = *(p+3) - '0';
+		if(ns < 0 || ns > 9) { /* not a digit */
+		    ns = asInteger(GetOption1(install("digits.secs")));
+		    if(ns == NA_INTEGER) ns = 0;
+		    nused = 3;
 		}
-#ifdef USE_INTERNAL_MKTIME
-		R_strftime(buff, 256, buf2, &tm);
-#else
-		strftime(buff, 256, buf2, &tm);
-#endif
-		if(UseTZ) {
-		    if(LENGTH(x) >= 10) {
-			const char *p = CHAR(STRING_ELT(VECTOR_ELT(x, 9), i));
-			if(strlen(p)) {strcat(buff, " "); strcat(buff, p);}
-		    } else if(!isNull(tz)) {
-			int ii = 0;
-			if(LENGTH(tz) == 3) {
-			    if(tm.tm_isdst > 0) ii = 2;
-			    else if(tm.tm_isdst == 0) ii = 1;
-			    else ii = 0; /* Use base timezone name */
-			}
-			const char *p = CHAR(STRING_ELT(tz, ii));
-			if(strlen(p)) {strcat(buff, " "); strcat(buff, p);}
-		    }
+		if(ns > 6) ns = 6;
+		if(ns > 0) {
+		    /* truncate to avoid nuisances such as PR#14579 */
+		    double s = secs, t = Rexp10((double) ns);
+		    s = ((int) (s*t))/t;
+		    sprintf(p2, "%0*.*f", ns+3, ns, s);
+		    strcat(buf2, p+nused);
+		} else {
+		    strcat(p2, "%S");			/* #nocov */
+		    strcat(buf2, p+nused);		/* #nocov */
 		}
-		SET_STRING_ELT(ans, i, mkChar(buff));
 	    }
+	    // The overflow behaviour is not determined by C99.
+	    // We assume truncation, and ensure termination.
+#ifdef USE_INTERNAL_MKTIME
+	    R_strftime(buff, 256, buf2, &tm);
+#else
+	    strftime(buff, 256, buf2, &tm);
+#endif
+	    buff[256] = '\0';
+	    // Now assume tzone abbreviated name is < 40 bytes,
+	    // but they are currently 3 or 4 bytes.
+	    if(UseTZ) {
+		if(have_zone) {
+		    const char *p = CHAR(STRING_ELT(VECTOR_ELT(x, 9), i%n));
+		    if(strlen(p)) {strcat(buff, " "); strcat(buff, p);}
+		} else if(!isNull(tz)) {		/* #nocov start */
+		    int ii = 0;
+		    if(LENGTH(tz) == 3) {
+			if(tm.tm_isdst > 0) ii = 2;
+			else if(tm.tm_isdst == 0) ii = 1;
+			else ii = 0; /* Use base timezone name */
+		    }
+		    const char *p = CHAR(STRING_ELT(tz, ii));
+		    if(strlen(p)) {strcat(buff, " "); strcat(buff, p);}
+		}					/* #nocov end */
+	    }
+	    SET_STRING_ELT(ans, i, mkChar(buff));
 	}
     }
     if(settz) reset_tz(oldtz);
@@ -1059,15 +1078,15 @@ SEXP Rstrptime(SEXP xarg, SEXP sformatarg, SEXP stzarg) {
     /* if(!isString((x = CAR(args)))) */
     /*     error(_("invalid '%s' argument"), "x"); */
     if (!isString(x))
-        error("invalid '%s' argument", "x"); 
+        error("invalid '%s' argument", "x"); 				/* #nocov */
     /* if(!isString((sformat = CADR(args))) || XLENGTH(sformat) == 0) */
     /*     error(_("invalid '%s' argument"), "x"); */
     if (!isString(sformat) || XLENGTH(sformat) == 0) 
-        error("invalid '%s' argument", "sformat"); 
+        error("invalid '%s' argument", "sformat");  			/* #nocov */
     /* if(!isString((stz = CADDR(args))) || LENGTH(stz) != 1) */
     /*     error(_("invalid '%s' value"), "tz"); */
     if (!isString(stz) || LENGTH(stz) != 1) 
-         error("invalid '%s' value", "tz"); 
+         error("invalid '%s' value", "tz");  				/* #nocov */
     tz = CHAR(STRING_ELT(stz, 0)); 
     if (strlen(tz) == 0) {
 	/* do a direct look up here as this does not otherwise
@@ -1089,14 +1108,14 @@ SEXP Rstrptime(SEXP xarg, SEXP sformatarg, SEXP stzarg) {
 
     // in case this gets changed by conversions.
     if (isgmt) {
-	PROTECT(tzone = mkString(tz));
+	PROTECT(tzone = mkString(tz)); 					/* #nocov */
     } else if(strlen(tz)) {
 	PROTECT(tzone = allocVector(STRSXP, 3));
 	SET_STRING_ELT(tzone, 0, mkChar(tz));
 	SET_STRING_ELT(tzone, 1, mkChar(R_tzname[0]));
 	SET_STRING_ELT(tzone, 2, mkChar(R_tzname[1]));
 
-    } else PROTECT(tzone); // for balance
+    } else PROTECT(tzone); // for balance 				/* #nocov */
 
     n = XLENGTH(x); m = XLENGTH(sformat);
     if(n > 0) N = (m > n) ? m : n; else N = 0;
@@ -1142,7 +1161,7 @@ SEXP Rstrptime(SEXP xarg, SEXP sformatarg, SEXP stzarg) {
 	    if(tm.tm_mday == 0) tm.tm_mday = NA_INTEGER;
 	    if(tm.tm_mon == NA_INTEGER || tm.tm_mday == NA_INTEGER
 	       || tm.tm_year == NA_INTEGER)
-		glibc_fix(&tm, &invalid);
+		glibc_fix(&tm, &invalid); 				/* #nocov */
 	    tm.tm_isdst = -1;
 	    if (offset != NA_INTEGER) {
 #ifdef HAVE_TM_GMTOFF
@@ -1152,12 +1171,12 @@ SEXP Rstrptime(SEXP xarg, SEXP sformatarg, SEXP stzarg) {
 		   so all we can do is to convert to time_t,
 		   adjust and convert back */
 		double t0;
-		memcpy(&tm2, &tm, sizeof(stm));
+		memcpy(&tm2, &tm, sizeof(stm)); 			/* #nocov start */
 		t0 = mktime0(&tm2, 0);
 		if (t0 != -1) {
 		    t0 -= offset; /* offset = -0800 is Seattle */
 		    ptm = localtime0(&t0, 1-isgmt, &tm2);
-		} else invalid = 1;
+		} else invalid = 1; 					/* #nocov end */
 	    } else {
 		/* we do want to set wday, yday, isdst, but not to
 		   adjust structure at DST boundaries */
@@ -1232,7 +1251,7 @@ SEXP D2POSIXlt(SEXP argsxp)
 	    if (day >= 0)
 		for ( ; day >= (tmp = days_in_year(y)); day -= tmp, y++);
 	    else
-		for ( ; day < 0; --y, day += days_in_year(y) );
+		for ( ; day < 0; --y, day += days_in_year(y) ); 	/* #nocov */
 
 	    y = tm.tm_year = y - 1900;
 	    tm.tm_yday = day;
@@ -1247,7 +1266,7 @@ SEXP D2POSIXlt(SEXP argsxp)
 	    tm.tm_isdst = 0; /* no dst in GMT */
 
 	    valid = 1;
-	} else valid = 0;
+	} else valid = 0; 						/* #nocov */
 	makelt(&tm, ans, i, valid, 0.0);
     }
     setAttrib(ans, R_NamesSymbol, ansnames);
@@ -1275,7 +1294,7 @@ SEXP POSIXlt2D(SEXP sxparg)
     /* PROTECT(x = duplicate(CAR(args))); */
     PROTECT(x = duplicate(sxparg)); 
     if(!isVectorList(x) || LENGTH(x) < 9) 
-        error("invalid '%s' argument", "x"); 
+        error("invalid '%s' argument", "x");  				/* #nocov */
 
     for(int i = 3; i < 6; i++)
 	if((nlen[i] = XLENGTH(VECTOR_ELT(x, i))) > n) n = nlen[i];
@@ -1283,9 +1302,9 @@ SEXP POSIXlt2D(SEXP sxparg)
     if(n > 0) {
 	for(int i = 3; i < 6; i++)
 	    if(nlen[i] == 0)
-		error("zero-length component in non-empty \"POSIXlt\" structure");
+		error("zero-length component in non-empty \"POSIXlt\" structure");  /* #nocov */
 	if(nlen[8] == 0)
-	    error("zero-length component in non-empty \"POSIXlt\" structure");
+	    error("zero-length component in non-empty \"POSIXlt\" structure");      /* #nocov */
     }
     /* coerce relevant fields to integer */
     for(int i = 3; i < 6; i++)
@@ -1301,7 +1320,7 @@ SEXP POSIXlt2D(SEXP sxparg)
 	tm.tm_isdst = 0;
 	if(tm.tm_mday == NA_INTEGER || tm.tm_mon == NA_INTEGER ||
 	   tm.tm_year == NA_INTEGER || validate_tm(&tm) < 0)
-	    REAL(ans)[i] = NA_REAL;
+	    REAL(ans)[i] = NA_REAL;  					/* #nocov */
 	else {
 	    /* -1 must be error as seconds were zeroed */
 	    double tmp = mktime00(&tm);
@@ -1314,3 +1333,78 @@ SEXP POSIXlt2D(SEXP sxparg)
     UNPROTECT(3);
     return ans;
 }
+
+/* Convert POSIXct to Date
+ *
+ * This is a combination of asPOSIXlt and POSIXlt2D that avoids some overhead
+ * when as.Date() is called on a POSIXct object with a non-UTC timezone. It
+ * avoids the intermediate (memory-inefficient) allocation to POSIXlt by
+ * re-using one tm struct instance for all POSIXct elements.
+ *
+ * Author: RCore for asPOSIXlt and POSIXlt2D.
+ *         Joshua Ulrich for the mashup.
+ */
+SEXP POSIXct2D(SEXP argsxp, SEXP tzarg)
+{
+    SEXP stz, x = argsxp, klass;
+    int isgmt = 0, valid = 0, settz = 0;
+    char oldtz[1001] = "";
+    const char *tz = NULL;
+
+    if (!isString((stz = tzarg)) || LENGTH(stz) != 1)
+        error("invalid '%s' value", "tz");  			/* #nocov */
+    tz = CHAR(STRING_ELT(stz, 0));
+    if(strlen(tz) == 0) {
+	/* do a direct look up here as this does not otherwise
+	   work on Windows */
+	char *p = getenv("TZ");  				/* #nocov start */
+	if(p) {
+	    stz = mkString(p); /* make a copy */
+	    tz = CHAR(STRING_ELT(stz, 0));			/* #nocov end */
+	}
+    }
+    PROTECT(stz); /* it might be new */
+    if(strcmp(tz, "GMT") == 0  || strcmp(tz, "UTC") == 0) isgmt = 1;
+    if(!isgmt && strlen(tz) > 0) settz = set_tz(tz, oldtz);
+#ifdef USE_INTERNAL_MKTIME
+    else R_tzsetwall(); // to get the system timezone recorded
+#else
+    tzset();
+#endif
+
+    R_xlen_t n = XLENGTH(x);
+
+    SEXP date_ans = PROTECT(allocVector(INTSXP, n));
+    int* date_ans_ = INTEGER(date_ans);
+    double* x_ = REAL(x);
+
+    for(R_xlen_t i = 0; i < n; i++) {
+	stm dummy, *ptm = &dummy;
+	double d = x_[i];
+	if(R_FINITE(d)) {
+	    ptm = localtime0(&d, 1 - isgmt, &dummy);
+	    /* in theory localtime/gmtime always return a valid
+	       struct tm pointer, but Windows uses NULL for error
+	       conditions (like negative times). */
+	    valid = (ptm != NULL);
+	} else {
+            valid = 0; 					 /* #nocov */
+        }
+
+        if(valid) {
+	    ptm->tm_sec = ptm->tm_min = ptm->tm_hour = 0;
+	    ptm->tm_isdst = 0;
+	    /* -1 must be error as seconds were zeroed */
+	    double tmp = mktime00(ptm);
+	    date_ans_[i] = (tmp == -1) ? NA_INTEGER : tmp/86400;
+        } else {
+	    date_ans_[i] = NA_INTEGER;			  /* #nocov */
+	}
+    }
+
+    if(settz) reset_tz(oldtz);
+    PROTECT(klass = mkString("Date"));
+    classgets(date_ans, klass);
+    UNPROTECT(3);
+    return date_ans;
+}
